from typing import Any

from pydantic import BaseModel, Field


class Locator(BaseModel):
    state_class_locator: str

    params: dict[str, Any] = Field(default_factory=dict)

    class Config:
        allow_mutation = True
        validate_assignment = True
        validate_all = True  # default values should be validated too
        extra = 'forbid'

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Like common pydantic.BaseModel init method with support of positional argument `state_class_locator`.

        Full compatabilite with pydantic.BaseModel.__init__ method is provided.
        """
        if len(args) == 1:
            super().__init__(state_class_locator=args[0], **kwargs)
        else:
            super().__init__(*args, **kwargs)


class FrozenLocator(Locator):
    class Config:
        allow_mutation = False



from functools import cached_property
from typing import Any

from pydantic import BaseModel, Field

from .locators import Locator, FrozenLocator


class BaseState(BaseModel):
    """Base class for all states of state machine.

    Fill free to inherite custom state class from BaseState with adding new attributes to Pydantic model scheme.
    """

    state_class_locator: str = Field(
        description='Path-like string specifies how to find required State class. '
                    'Will be initialized by router on state instance creation.',
    )

    class Config:
        allow_mutation = False
        validate_all = True  # default values should be validated too
        extra = 'ignore'
        keep_untouched = (cached_property,)

    def enter_state(self) -> Locator | None:
        """Run any custom logic on state enter.

        Can return state object to force state machine switching to another state.
        """
        pass

    async def aenter_state(self) -> Locator | None:
        """Run any custom logic on state enter.

        Can return state object to force state machine switching to another state.
        """
        pass

    def exit_state(self, state_class_transition: bool) -> None:
        """Run any custom logic on state exit.

        State machine switching to another state is not available from this method.
        """
        pass

    async def aexit_state(self, state_class_transition: bool) -> None:
        """Run any custom logic on state exit.

        State machine switching to another state is not available from this method.
        """
        pass

    def process(self, event: Any) -> Locator | None:
        """Run any custom logic to process event.

        Can return state object to force state machine switching to another state.
        """
        pass

    async def aprocess(self, event: Any) -> Locator | None:
        """Run any custom logic to process event.

        Can return state object to force state machine switching to another state.
        """
        pass

    @cached_property
    def locator(self) -> Locator:
        return FrozenLocator(
            state_class_locator=self.state_class_locator,
            params=self.dict(
                exclude={'state_class_locator'},
                by_alias=True,
                exclude_defaults=True,  # Make locators shorter
            ),
        )



from .router import Router, Route, StateDecoratorType  # noqa F401
from .states import BaseState  # noqa F401
from .exceptions import (  # noqa F401
    LocatorError,
    NotFoundStateClassLocatorError,
    LocatorParamsError,
    TooLongTransitionError,
    DetachedCrawlerError,
)
from .locators import Locator, FrozenLocator  # noqa F401
from .sync_crawler import Crawler  # noqa F401



from typing import Type, Callable, Sequence

from pydantic import ValidationError, validate_arguments, BaseModel

from .locators import Locator
from .exceptions import NotFoundStateClassLocatorError, LocatorParamsError
from .states import BaseState
from .state_class_locator_validators import validate_state_class_locator, StateClassLocatorValidator

StateDecoratorType = Callable[[Type[BaseState]], Type[BaseState]]


class Route(BaseModel):
    state_class_locator: str
    state_class: Type[BaseState]
    title: str = ''

    class Config:
        allow_mutation = False
        validate_all = True  # default values should be validated too


class Router(dict[str, Route]):
    """Index of registered state classes."""

    decorators: tuple[StateDecoratorType, ...]
    state_class_locator_validators: tuple[StateClassLocatorValidator]

    @validate_arguments
    def __init__(
        self,
        decorators: Sequence[StateDecoratorType] | None = None,
        state_class_locator_validators: Sequence[Callable[[str], str]] | None = (validate_state_class_locator, ),
    ):
        self.decorators = tuple(decorators) if decorators else tuple()
        self.state_class_locator_validators = state_class_locator_validators  # type: ignore

    @validate_arguments
    def register(self, state_class_locator: str, *, title: str = '') -> Type[BaseState]:
        """Register a State with specified locator."""
        @validate_arguments
        def register_state_class(state_class: Type[BaseState]) -> Type[BaseState]:
            wrapped_state_class = state_class
            for decorator in reversed(self.decorators):
                wrapped_state_class = decorator(wrapped_state_class)

            cleaned_state_class_locator = self._validate_state_class_locator(state_class_locator)

            route = Route(
                state_class_locator=cleaned_state_class_locator,
                state_class=wrapped_state_class,
                title=title,
            )
            self[route.state_class_locator] = route

            return wrapped_state_class
        return register_state_class

    @validate_arguments
    def create_state(self, locator: Locator) -> BaseState:
        """Create new serializable State."""
        cleaned_state_class_locator = self._validate_state_class_locator(locator.state_class_locator)

        route = self.get(cleaned_state_class_locator, None)

        if not route:
            raise NotFoundStateClassLocatorError(
                f'Unknown state class locator {locator.state_class_locator!r} with '
                f'normalized value {cleaned_state_class_locator!r}',
            )

        try:
            return route.state_class.parse_obj(locator.params | {
                'state_class_locator': cleaned_state_class_locator,
            })
        except ValidationError as error:
            raise LocatorParamsError(f'Can`t create state for locator {locator}') from error

    def _validate_state_class_locator(self, value: str) -> str:
        for validate in self.state_class_locator_validators:
            value = validate(value)

        return value



import re
import string
from typing import Callable

STATE_CLASS_LOCATOR_REGEXP = re.compile(r"^/([a-z0-9_\-]+/)*$")


StateClassLocatorValidator = Callable[[str], str]


def validate_state_class_locator(value: str) -> str:
    """Проверяет Локатор класса состояний на соответствие принятому соглашению по наименованию.

    Проверяет Локатор класса состояний по максимально строгим правилам:
    - только ASCII символы
    - только нижний регистр
    - начинается и заканчивается слэшом `/`
    - без пробельных символов

    Функция только следит за соответствием, но не пытается сама нормализовать значение.
    """
    value = str(value)

    if not value.startswith('/'):
        raise ValueError(
            f'Wrong state class locator string format {value!r}. Leading slash symbol `/` is absent.',
        )

    if not value.endswith('/'):
        raise ValueError(
            f'Wrong state class locator string format {value!r}. Trailing slash symbol `/` is absent.',
        )

    if ' ' in value:
        raise ValueError(
            f'Wrong state class locator string format {value!r}. '
            'Whitespace symbols are found. Use dashes and underscores instead.',
        )

    found_uppercase_symbols = set(value) & set(string.ascii_uppercase)
    if set(value) & set(string.ascii_uppercase):
        raise ValueError(
            f'Wrong state class locator string format {value!r}. '
            f'Uppercase symbols are found: {found_uppercase_symbols!r}.',
        )

    allowed_symbols = f'{string.ascii_lowercase}{string.digits}_-/'
    prohibited_symbols = set(value) - set(allowed_symbols)

    if prohibited_symbols:
        raise ValueError(
            f'Wrong state class locator string format {value!r}. '
            'Prohibited symbols are found: {prohibited_symbols!r}.',
        )

    if not STATE_CLASS_LOCATOR_REGEXP.match(value):
        raise ValueError(
            f'Wrong state class locator string format {value!r}. Check out STATE_CLASS_LOCATOR_REGEXP.',
        )

    return value



from itertools import count
import logging
from typing import final, Any
from typing import cast

from pydantic import validate_arguments

from .states import BaseState
from .router import Router, Locator
from .exceptions import (
    TooLongTransitionError,
    LocatorError,
    DetachedCrawlerError,
)

logger = logging.getLogger('yostate')


@final
class Crawler:
    router: Router
    current_state: BaseState | None = None
    max_transition_length: int = 20

    @validate_arguments
    def __init__(
        self,
        router: Router,
        max_transition_length: int = max_transition_length,
    ):
        """Инициализирует краулер.

        Аргумент `max_transition_length` настраивает защиту от зацикливания. Он ограничивает максимальную длину
        непрерывной цепочки переходов между состояниями.
        """
        self.router = router
        self.max_transition_length = max_transition_length

    @validate_arguments
    def restore(self, locator: Locator, ignore_errors: bool = False) -> None:
        """Восстанавливает положение краулера в прежнем состоянии.

        Используйте метод `restore` чтобы восстановить положение краулера в том состоянии, где он ранее прервал
        свою работу.
        Вызов метода `restore` отличается от `switch_to` тем, что не приводит к запуску кода в методе
        `BaseState.enter_state` и не запускает процесс переходов по состояниям.
        """
        try:
            self.current_state = self.router.create_state(locator)
        except LocatorError:
            if not ignore_errors:
                raise
            logger.warning('Crawler restore failed for locator %s', locator)

    def detach(self) -> None:
        self.current_state = None

    def process(self, event: Any) -> None:
        """Обрабатывает поступившее событие."""
        if self.attached:
            current_state = cast(BaseState, self.current_state)
        else:
            raise DetachedCrawlerError('Crawler is not attached yet')

        next_locator = current_state.process(event=event)

        if next_locator:
            self.switch_to(next_locator)

    @validate_arguments
    def switch_to(self, locator: Locator) -> None:  # noqa CCR001
        """Переводит краулер в новое состояние и следует далее по цепочке переходов до упора.

        В краулер встроена защиты от зацикливания. Она ограничивает максимальную длину цепочки переходов.
        """
        next_state = self.router.create_state(locator)

        counter = count(1)

        prev_state = self.current_state

        for transition_length in counter:
            if transition_length > self.max_transition_length:
                raise TooLongTransitionError(
                    f'Transition length limit of {self.max_transition_length} is exceeded.',
                )

            logger.debug(
                'State %s → %s.',
                prev_state and prev_state.state_class_locator,
                next_state.state_class_locator,
            )
            logger.debug('    Old: %s', prev_state)
            logger.debug('    New: %s', next_state)

            state_class_transition = type(prev_state) != type(next_state)

            if prev_state:
                prev_state.exit_state(state_class_transition=state_class_transition)

            next_next_locator = next_state.enter_state()
            if not next_next_locator:
                break

            next_next_state = self.router.create_state(next_next_locator)
            prev_state, next_state = next_state, next_next_state

        self.current_state = next_state

    @property
    def attached(self) -> bool:
        return bool(self.current_state)



class LocatorError(RuntimeError):
    pass


class NotFoundStateClassLocatorError(LocatorError):
    pass


class LocatorParamsError(LocatorError):
    pass


class TooLongTransitionError(RuntimeError):
    pass


class DetachedCrawlerError(RuntimeError):
    pass
